import fiona
from pyproj import Proj, transform
from shapely.geometry import mapping, shape
import logging as log
from functools import partial
from fiona.crs import from_string

def transform_coords(func, records):
    """
    Transform record geometry coordinates using the provided function.
    .. http://sgillies.net/blog/1125/fiona-pyproj-shapely-in-a-functional-style/
    """
    for rec in records:
        try:
            assert rec['geometry']['type'] == "LineString"
            coords = func(*zip(*rec["geometry"]["coordinates"]))
            rec['geometry']['coordinates'] = zip(*coords)
        except Exception, e:
            # Writing untransformed features to a different shapefile
            # is another option.
            log.exception("Error transforming record %s:", rec['id'])
        yield rec

class Transformation(object):
    def __init__(self, source, sink):
        self.source = Proj(source)
        self.sink = Proj(sink)

    def transform(self, records):
        func = partial(transform, self.source, self.sink)
        return transform_coords(func, records)

    def reversed(self, records):
        func = partial(transform, self.sink, self.source)
        return transform_coords(func, records)

def get_crs(argument):
    try:
        with open(argument, "r") as f:
            argument = f.read()
    except IOError, err:
        pass
    if argument[0] != "+":
        spatial_ref = osr.SpatialReference()
        spatial_ref.ImportFromWkt(argument)
        argument = spatial_ref.ExportToProj4()
    return from_string(argument)



from . import Regression
from ..tests import random_plane,scattered_plane
import numpy as N

def test_coordinates():
	"""Tests coordinate length"""
	plane,coefficients = random_plane()
	fit = Regression(plane)
	assert N.column_stack(plane).shape[0] == fit.C.shape[0]

def test_regression():
	"""Make sure we can fit a simple plane"""
	plane,coefficients = random_plane()
	fit = Regression(plane)
	assert N.allclose(fit.coefficients(), coefficients)

def test_covariance():
	"""Make sure we don't get empty covariance matrices"""
	plane,coefficients = scattered_plane()
	fit = Regression(plane)
	for i in fit.covariance_matrix().flatten():
		assert i != 0



from __future__ import division
import numpy as N

import logging
log = logging.getLogger(__name__)

def add_ones(a):
    """Adds a column of 1s at the end of the array"""
    arr = N.ones((a.shape[0],a.shape[1]+1))
    arr[:,:-1] = a
    return arr

def matrix_squared(array):
    """Returns a pseudo-squared matrix by solving X^T.X"""
    return N.dot(array.transpose(),array)

class Regression(object):
    def __init__(self, coordinates):
        """
        Solves plane equation Z = a(C[0])+b(C[1])+c by solving linear equation Cm=Z where m=[a,b,c]
        C and Z are the dependent and independent variables, respectively.
        """
        coordinates = N.array(coordinates)
        if len(coordinates[0]) > 3:
            coordinates = N.swapaxes(coordinates, 0, 1)

        self.C = add_ones(coordinates[:,0:2])
        self.Z = coordinates[:,-1]
        self.__coefficients__ = None
        self.__covariance__ = None

        self.nobs = self.Z.shape[0]                     # number of observations
        self.ncoef = self.C.shape[1]                    # number of coef.
        self.df_e = self.nobs - self.ncoef              # degrees of freedom, error 
        self.df_r = self.ncoef - 1                      # degrees of freedom, regression 
        self.n,self.k = self.C.shape

    def coefficients(self):
        if self.__coefficients__ is not None:
            return self.__coefficients__
        self.xTx = N.linalg.inv(matrix_squared(self.C))
        self.xy = N.dot(self.C.transpose(),self.Z)
        self.__coefficients__ = N.dot(self.xTx,self.xy)
        return self.__coefficients__

    def predictions(self):
        return self.C[:,0], self.C[:,1], N.dot(self.C,self.coefficients())

    def residuals(self):
        return self.Z - self.predictions()[2]

    def covariance_matrix(self):
        if self.__covariance__ is not None:
            return self.__covariance__
        if self.n == self.k:
            # solution is fully defined (i.e. only three points)
            self.__covariance__ = N.zeros((self.n,self.k))
        else:
            #rTr = matrix_squared(self.residuals())
            e = self.residuals()
            sse = N.dot(e,e)/self.df_e #SSE
            self.__covariance__ = N.dot(sse,self.xTx)
        return self.__covariance__

    def standard_errors(self):
        """Gets the standard errors of the coefficients"""
        return N.sqrt(N.diagonal(self.covariance_matrix()))



import json

def dump_features(features, file, float_format=""):
	class PrettyFloat(float):
		def __repr__(self):
			return '%.2f' % self

	def pretty_floats(obj):
		if isinstance(obj, float):
			return PrettyFloat(obj)
		elif isinstance(obj, dict):
			return dict((k, pretty_floats(v)) for k, v in obj.items())
		elif isinstance(obj, (list, tuple)):
			return map(pretty_floats, obj)             
		return obj

	struct = {
		"type": "FeatureCollection",
		"features": list(features)
	}
	json.dump(pretty_floats(struct), file)


import numpy as N

def spherical(coordinates):
    """No error is propagated"""
    x,y,z = coordinates
    r = N.sqrt(x**2+y**2+z**2)
    theta = N.arccos(z/r)
    phi = N.arctan2(y,x)
    return r,theta,phi

def cartesian(spherical):
	r,theta,phi = spherical
	z = r*N.cos(theta)
	zc = r*N.sin(theta)
	x = zc*N.sin(phi)
	y = zc*N.cos(phi)
	return x,y,z

def centered(coordinates):
	try:
		assert coordinates.shape[1] == 3
	except Exception:
		coordinates = N.column_stack(coordinates)
	
	means = N.mean(coordinates,axis=0,keepdims=True)
	n = coordinates - means
	return tuple(a.flatten() for a in n.transpose())


from __future__ import division
import numpy as N
from ..regression import Regression
from ..coordinates import spherical, cartesian

def spherical_plane(r,theta,coefficients=[1,1]):
	c = coefficients
	return r,theta,c[0]*N.sin(theta+c[1])

def strike_dip_plane(strike,dip):
	pass

def make_plane(x, y, coefficients):
	"""z = a x + b y + c
		scatter: percentage scatter (standard deviation) in final values (0 for exact plane)
	"""
	try:
		assert len(x) == len(y)
	except AssertionError:
		raise ValueError("X and Y must be the same length.")

	c = coefficients
	return c[0]*x+c[1]*y+c[2]

def randomize(plane, scatter=.05):
	return tuple(i*(1+N.random.randn(len(i))*scatter) for i in plane)

def random_plane(n=1000, limit=100):
    x,y = N.random.uniform(-limit,limit,(2,n))
    coefficients = N.random.randn(3)*N.array([1,1,limit/2])
    z = make_plane(x, y, coefficients)
    return (x,y,z),coefficients

def scattered_plane(*args,**kwargs):
	scatter = kwargs.pop("scatter",0.5)
	plane, coefficients = random_plane(*args,**kwargs)
	return randomize(plane,scatter), coefficients

def test_coordinate_transforms(n_points=1000):
	xyz = N.random.randn(3, n_points)
	for i,j in zip(xyz, cartesian(spherical(xyz))):
		assert i.all() == j.all()

def test_random_plane(n_points=1000):
	"""Test that we can fit a simple plane in cartesian space"""

	n = 1e5 # A generically large number
	x,y = N.random.uniform(-n,n,(2,n_points))
	coefficients = N.random.randn(3)*10
	z = make_plane(x, y, coefficients)

	fitted_coefficients = Regression((x,y,z)).coefficients()
	print coefficients, fitted_coefficients
	for i, j in zip(coefficients,fitted_coefficients):
		assert i - j < 0.0001



import matplotlib.pyplot as P
from mplstereonet.stereonet_math import line, pole
import numpy as N
from matplotlib.patches import Polygon

def trend_plunge(orientation, *args, **kwargs):
	ax = kwargs.pop("ax",P.gca())
	levels = kwargs.pop("levels",[1])
	kwargs["linewidth"] = 0

	a = kwargs.pop("alpha",0.7)
	if len(a) != len(levels):
		a = [a]*len(levels)

	for i,level in enumerate(levels):
		el = map(N.degrees,orientation.error_ellipse(vector=True, level=level))
		lat,lon = line(el[1], el[0])
		e = Polygon(zip(lat,lon), alpha=a[i], **kwargs)
		ax.add_artist(e)

def normal(orientation, *args, **kwargs):
	ax = kwargs.pop("ax",P.gca())
	levels = kwargs.pop("levels",[1])
	kwargs["linewidth"] = 0

	a = kwargs.pop("alpha",0.7)
	if len(a) != len(levels):
		a = [a]*len(levels)

	for i,level in enumerate(levels):
		el = map(N.degrees,orientation.error_ellipse(vector=True, level=level))
		lat,lon = line(90-el[1], 180+el[0])
		e = Polygon(zip(lat,lon), alpha=a[i], **kwargs)
		ax.add_artist(e)

def setup_figure(*args, **kwargs):
	fig = P.figure(*args, **kwargs)
	ax = fig.add_subplot(111, projection='stereonet')
	return fig,ax



from __future__ import division
import numpy as N

def ellipse(center,covariance_matrix,level=1, n=1000):
    """Returns error ellipse in slope-azimuth space"""
    # singular value decomposition
    U, s, rotation_matrix = N.linalg.svd(covariance_matrix)
    # semi-axes (largest first)

    saxes = N.sqrt(s)*level ## If the _area_ of a 2s ellipse is twice that of a 1s ellipse
    # If the _axes_ are supposed to be twice as long, then it should be N.sqrt(s)*width

    u = N.linspace(0, 2*N.pi, n)
    data = N.column_stack((saxes[0]*N.cos(u), saxes[1]*N.sin(u)))
    # rotate data
    return N.dot(data, rotation_matrix)+ center



from __future__ import division
from ..regression import Regression
import numpy as N
from scipy.linalg import eig
from ..error.ellipse import ellipse

def rotation(angle):
	"""Rotation about the Z axis (in the XY plane)"""
	return N.array([[N.cos(angle),-N.sin(angle),0],
		[N.sin(angle), N.cos(angle),0],
		[0           , 0           ,1]])

def axes(matrix):
	"""
	Computes the ellipse axes lengths for a covariance matrix
	"""
	return N.sqrt(eig(matrix[0:2,0:2])[0])

class Orientation(object):
	def __init__(self, coordinates):
		self.fit = Regression(coordinates)

		values = self.fit.coefficients()
		val = values[0]**2+values[1]**2
		self.azimuth = -N.arctan2(-values[1], -values[0])
		self.rotation = rotation(self.azimuth)
		self.coefficients = N.dot(self.rotation,values)
		self.slope = N.arctan(-self.coefficients[0])

	def covariance_matrix(self):
		return N.dot(self.rotation,self.fit.covariance_matrix())

	def standard_errors(self):
		return N.sqrt(N.diagonal(self.covariance_matrix()))

	def strike_dip(self, uncertainties=False):
		c = tuple(N.degrees(i) for i in (self.azimuth,self.slope))
		if uncertainties:
			return c,self.errors()
		return c

	def dip_direction(self, uncertainties=False):
		s,d = self.strike_dip()
		s+=90
		if uncertainties:
			return (s,d),self.errors()
		return (s,d)

	def gradient(self, uncertainties=False):
		return self.azimuth,self.slope

	def errors(self):
		return tuple(N.degrees(i) for i in self.standard_errors()[:2])

	def error_ellipse(self, spherical=True, vector=False, level=1):
		e = ellipse(tuple(self.coefficients[:2]), self.covariance_matrix()[:2,:2], level=level)
		if spherical:
			slope = N.arctan(-e[:,0])
			azimuth = self.azimuth + N.arctan2(-e[:,1],-e[:,0])
			if vector:
				azimuth = azimuth + N.pi/2
			return (azimuth,slope)
		return (e[:,1],e[:,0])



import os
from setuptools import setup, find_packages
here = os.path.abspath(os.path.dirname(__file__))

install_requires = [
    'numpy',
    'scipy',
    'matplotlib',
    'fiona',
    'shapely',
    'pyproj',
    'mplstereonet'
    ]

setup(
    name='Attitude',
    version=0.1,
    description="Attitude computes the orientations of planes from point data.",
    license='MIT',
    keywords='gis data computation fitting statistics vector science geology',
    author='Daven Quinn',
    author_email='dev@davenquinn.com',
    maintainer='Daven Quinn',
    maintainer_email='dev@davenquinn.com',
    url='http://github.com/davenquinn/Attitude',
    install_requires=install_requires,
    tests_require=['nose'],
    test_suite='nose.collector',
    packages=find_packages(),
    package_dir={'attitude':'attitude'},
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: GIS',
    ],
)


import numpy as N
from ..orientation import Orientation
import json
from shapely.geometry import shape
from ..coordinates import centered


def fit_planes(features):
    for i,feature in enumerate(features):
        coords = N.array(shape(feature["geometry"]))
        fit = Orientation(centered(coords))
        yield fit




from . import Orientation
import numpy as N
import nose
import functools
from ..tests import random_plane, scattered_plane


simple_cases = [
    ([(1,0,1),(0,0,1),(0,1,0)], [-90, 45], "North-dipping"),
    ([(5,0,-5), (0,5,0), (0,0,0)], [0,45], "East-dipping"),
    ([(1,0,0),(1,1,1), (2,1,1)], [90,45], "South-dipping"),
    ([(0,2,0),(1,1,1), (0,0,0)], [180,45], "West-dipping"),
    ([(0,2,1),(1,1,0), (0,0,1)], [0,45], "East-dipping (sloping up)"),
    ([(0,0,1),(1,0,0),(0,1,0)], [-45,N.degrees(N.arctan(N.sqrt(2)))], "Northeast-dipping")
]

edge_cases = [
    ([(0,0,0),(1,1,0), (0,2,0), (4,2,0)], [0,0], "Flat"),
    ([(1,0,0),(1,1,-1),(1,2,1)],[0,90], "Vertical, facing east")
]

def check_orientation(a):
    coords, sol, name = a
    print name
    sol = tuple(sol)
    res = Orientation(coords).strike_dip()
    print res, sol
    print ""
    assert res[0] == sol[0] and res[1] == sol[1]

def test_directions():
    """
    Tests cardinal directions for simple cases, to make sure everything conforms to right-hand rule.
    Does not test the standard deviation of these methods.
    """
    for a in simple_cases:
        yield check_orientation, a

def test_coordinates():
    """Tests to make sure we don't lose coordinates in the Orientation object"""
    plane,coordinates = random_plane()
    orient = Orientation(plane)
    assert len(plane[0]) == len(orient.fit.C)

def test_covariance():
    """Make sure we don't get empty covariance matrices in the Orientation object"""
    plane,coefficients = scattered_plane()
    fit = Orientation(plane)
    for i in fit.covariance_matrix().flatten():
        assert i != 0

def test_rotation_matrix():
    """Check that we have properly rotated our matrix to bring the along-strike direction into a coordinate plane."""
    plane,coefficients = scattered_plane()
    fit = Orientation(plane)
    assert N.allclose(fit.coefficients[1],0)
